/***************************************************************************
                                                                   
				DT2851 DRIVER                              
	Copyright (C) 1992..1999 Walter Fetter Lages <w.fetter@ieee.org> 

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    History:
    2.01 04.04.97  First Linux version.
    2.02 09.23.98  Created a local copy of personality.h to avoid kernel patch.
    2.20 11.19.99  Ported to kernel series 2.2.x.
  		   Ported back to the C language, to avoid problems with kernel
  		   includes.
  		   Eliminated local copy of personality.h.
    2.2.1 05.12.99 Changed version numbers to major.minor.revision.
		   Corrected some typing errors.
		   Added some man pages. Thanks to Primoz Peterlin 
		   <peterlin@biofiz.mf.uni-lj.si>
                                                                   
*****************************************************************************/

#define MODULE
#define __KERNEL__

#include <linux/module.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>

#include <linux/fs.h>
#include <linux/ioport.h>
#include <linux/sched.h>

#include <unistd.h>

#include <dt2851io.h>


/* just because I like it this way... */ 
#define outportw(port,data)	outw(data,port)
#define inportw(port)		inw(port)


/* support old naming sillyness */ 
#if LINUX_VERSION_CODE < 0x020100                                     
#define ioremap vremap
#define iounmap vfree                                                     
#endif

/* don't use #define. This way these paramenters are tunable:
   insmod dt2851 <paramenter>=<value> */

static int major=26;
static unsigned int baseio[4]={0x230,0x230,0x230,0x230};
static unsigned long basemem[4]={0xa00000,0xa00000,0xa00000,0xa00000};
static int irq[4]={15,15,15,15};

MODULE_PARM(major,"i");
MODULE_PARM(baseaio,"1-4i");
MODULE_PARM(basemem,"1-4l");
MODULE_PARM(irq,"1-4i");

enum MINOR
{
	BUF0_MINOR=	0,
	BUF1_MINOR=	1,
	ILUT0_MINOR=	2,
	ILUT1_MINOR=	3,
	ILUT2_MINOR=	4,
	ILUT3_MINOR=	5,
	ILUT4_MINOR=	6,
	ILUT5_MINOR=	7,
	ILUT6_MINOR=	8,
	ILUT7_MINOR=	9,
	OLUT0_MINOR=	10,
	OLUT1_MINOR=	11,
	OLUT2_MINOR=	12,
	OLUT3_MINOR=	13,
	OLUT4_MINOR=	14,
	OLUT5_MINOR=	15,
	OLUT6_MINOR=	16,
	OLUT7_MINOR=	17,
	CUR_MINOR=	18,
	CS_MINOR=	19
};

#define DT_IO_EXTENT 16
#define BUF_SIZE (512L*512L)
#define LUT_SIZE 256

/* DT2851 registers definitions */
enum REGISTERS
{
	INCSR1=	0x0000,	/* video input control/status register 1 */
	INCSR2=	0x0002,	/* video input control/status register 2 */
	OUTCSR=	0x0004,	/* video output control/status register */
	CURSOR=	0x0006,	/* cursor register */
	INDEX=	0x0008,	/* index  register */
	INLUT=	0x000a,	/* input look-up table entry register */
	REDGRN=	0x000c,	/* red and green output look-up table register */
	BLUE=	0x000e	/* blue output look-up table register */
};


/* video input control/status register 1 definitions */
enum CTRLREG1
{
	BUSY=	0x0080,
	DONEINT=0x0040,	/* interrupt on done */
	EXTTRG=	0x0020,	/* external trigger */
	FRPROC=	0x0010,	/* frame processor */
	ENSTOP=	0x0008,	/* enable stop */
	ISEL=	0x0007,	/* input LUT select */
	ISEL2=	0x0004,
	ISEL1=	0x0002,
	ISEL0=	0x0001
};

/* video input control/status register 2 definitions */
enum CTRLREG2
{
	BUFSEL=		0x0080,	/* buffer select */
	MODE=		0x0070, /* mode */
	FEEDBACK=	0x0000,
	VIDEOIN=	0x0010,
	LOADLUT=	0x0040,
	SLOWSCAN=	0x0050,
	PORTOUT=	0x0060,
	PORTIN=		0x0070,
	WP=		0x000f,	/* write protect */
	WP3=		0x0008,
	WP2=		0x0004,
	WP1=		0x0002,
	WP0=		0x0001
};


/* video output control/status register definitions */
enum VIDREG
{
	VSYNC=		0x8000,	/* vertical sync */
	SYNCINT=	0x4000,	/* interrupt on sync */
	FIELD=		0x2000,
	/* EXTTRG=	0x1000,	*/ /* This is a duplicate of EXTTRG in INCSR1 */
	/* RESERVED=	0x0800, */
	/* RESERVED=	0x0400, */
	/* RESERVED=	0x0200, */
	/* RESERVED=	0x0100, */
	DISP=		0x0080,
	CURS=		0x0040,	/* cursor */
	EXTTIM=		0x0020,	/* external timing */
	DISBUF=		0x0010,	/* diplay buffer select */
	OSEL=		0x000f,	/* output LUT select */
	OSEL3=		0x0008,
	OSEL2=		0x0004,
	OSEL1=		0x0002,
	OSEL0=		0x0001
};


/* cursor register definitions */
enum CURREG
{
	LINE=	0xff00,	/* line address/2 */
	PIXEL=	0x00ff	/* pixel address/2 */
};

/* index register definition */
#define LUTENTRY	0x00ff	/* select LUT entry */

/* input look-up table entry register definition */
#define LUTDATA		0x00ff	/* contains LUT data */

/* red and green output look-up table register definitions */
#define GREENDATA	0xff00	/* green LUT data */
#define REDDATA		0x00ff	/* red LUT data */

/* blue output look-up table register definition */
#define BLUEDATA	0x00ff	/* blue LUT data */


static unsigned int baseio[4];
static unsigned long basemem[4];
static struct
{
	int irq;
	struct task_struct *p;
} irq_data[4];

int getboard(struct file *file)
{
	return MINOR(file->f_dentry->d_inode->i_rdev)/(CS_MINOR-BUF0_MINOR+1);
}

int getdev(struct file *file)
{
	return MINOR(file->f_dentry->d_inode->i_rdev) % (CS_MINOR-BUF0_MINOR+1);
}

void irq_handler(int irq,void *dev_id,struct pt_regs *regs)
{
	int i;
	for(i=0; i < 4;i++)
	{
		if((irq==irq_data[i].irq) && (irq_data[i].p != NULL)) send_sig(SIGIO,irq_data[i].p,1);
	}
}

static void stop(int board)
{
	unsigned int incsr1=baseio[board]+INCSR1;
	unsigned short old=inportw(incsr1);
	outportw(incsr1,old | ENSTOP);
	while(inportw(incsr1) & BUSY);
}

static void clearilut(int board,int ilut)
{	
	unsigned int base=baseio[board];
	unsigned short incsr1=inportw(base+INCSR1);
	unsigned short incsr2;
	int i;

	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);
	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);

	outportw(base+INCSR1,ilut);
	
	for(i=0;i < LUT_SIZE;i++)
	{
		outportw(base+INDEX,i);

		switch(ilut)
		{
			case 0:
			{
				/* gray-scale */
				outportw(base+INLUT,i);
				break;
			}
			case 1:
			{
				/* invese gray-scale */
				outportw(base+INLUT,~i);
				break;
			}
			case 2:
			{	
				/* right-justified 4 MSBs */
				outportw(base+INLUT,i >> 4);
				break;
			}
			case 3:
			{
				/* left-justified 4 MSBs */
				outportw(base+INLUT,i & 0xf0);
				break;
			}
			case 4:
			{
				/* left-justified 6 MSBs */
				outportw(base+INLUT,i & 0xfc);
				break;
			}
			case 5:
			{
				/* left-justified 7 MSBs */
				outportw(base+INLUT,i & 0xfe);
				break;
			}
			case 6:
			{
				/* bi-level threshold */
				if(i > 0x80) outportw(base+INLUT,0xff);
				else outportw(base+INLUT,0x00);
				break;
			}
			case 7:
			{
				/* inverse bi-level threshod */
				if(i > 0x80) outportw(base+INLUT,0x00);
				else outportw(base+INLUT,0xff);
			}
		}
	}
	outportw(base+INCSR2,incsr2);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));
}

static void clearolut(int board,int olut)
{	
	unsigned int base=baseio[board];

	unsigned short incsr1=inportw(base+INCSR1);
	unsigned short incsr2;
	unsigned short outcsr;
	int i;

	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);
	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	outcsr=inportw(base+OUTCSR);
	incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);
	outportw(base+OUTCSR,olut);
	
	for(i=0;i < LUT_SIZE;i++)
	{
		outportw(base+INDEX,i);

		switch(olut)
		{
			case 0:
			{
				/* gray-scale */
				outportw(base+BLUE,i);
				outportw(base+REDGRN,i | (i << 8));
				break;
			}
			case 1:
			{
				/* inverse gray-scale */
				unsigned short o=~i & 0xff;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 2:
			{
				/* right-justified 4-bit image */
				unsigned short o=i << 4;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 3:
			{
				/* left-justified 4-bit image */
				unsigned short o=i & 0xf0;
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 4:
			{
				/* addition of two 4-bit images */
				unsigned short o1=i >> 4;
				unsigned short o2=i & 0x0f;
				unsigned short o=(o1+o2);
				if(o > 0xff) o=0xff;	/* saturation */
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 5:
			{
				/* subtraction of two 4-bit images */
				unsigned short o1=i >> 4;
				unsigned short o2=i & 0x0f;
				unsigned short o=(o1-o2) & 0xff;
				if(o & 0x10) o=0; /* negative saturation */
				outportw(base+BLUE,o);
				outportw(base+REDGRN,o | (o << 8));
				break;
			}
			case 6:
			{
				/* gray-scale image with color overlay */
				switch(i & 0x03)
				{
					case 0x00:
					{
						outportw(base+BLUE,i);
						outportw(base+REDGRN,i | (i << 8));
						break;
					}
					case 0x01:
					{	
						outportw(base+BLUE,0);
						outportw(base+REDGRN,REDDATA);
						break;
					}
					case 0x02:
					{
						outportw(base+BLUE,0);
						outportw(base+REDGRN,GREENDATA);
						break;
					}
					default:
					{
						outportw(base+BLUE,0xff);
						outportw(base+REDGRN,0x0000);
					}
				}
				break;
			}
			case 7:
			{
				/*  gray-scale image with red overlay */
				if(i & 1)
				{
					outportw(base+BLUE,0);
					outportw(base+REDGRN,REDDATA);
				}
				else
				{
					outportw(base+BLUE,i);
					outportw(base+REDGRN,i | (i << 8));
				}
				break;
			}
		}
	}
	outportw(base+INCSR2,incsr2);
	outportw(base+OUTCSR,outcsr);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));
}

static void clear(int board)
{
	unsigned int base=baseio[board];

	unsigned short outcsr=inportw(base+OUTCSR);
	unsigned short incsr1;
	unsigned short incsr2;
	int i;
	char *p;
	
	outportw(base+OUTCSR,outcsr & ~(SYNCINT | DISP | CURS | EXTTIM | DISBUF | OSEL));

	incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);


/* program look-up tables */

	do
	{
		incsr1=inportw(base+INCSR1);
	}while(incsr1 & BUSY);

	outcsr=inportw(base+OUTCSR);
	incsr2=inportw(base+INCSR2);
	outportw(base+INCSR2,LOADLUT);
	
	for(i=0;i < LUT_SIZE;i++)
	{
		unsigned short ilut;
		unsigned short olut;
		unsigned short o;
		unsigned short o1;
		unsigned short o2;

		ilut=inportw(base+INCSR1);
		ilut&=~ISEL;
		olut=inportw(base+OUTCSR);
		olut&=~OSEL;
		outportw(base+INDEX,i);

/* program input look-up table 0 with gray-scale */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i);

/* program output look-up table 0 with gray-scale */
		outportw(base+OUTCSR,olut++);
		outportw(base+BLUE,i);
		outportw(base+REDGRN,i | (i << 8));

/* program input look-up table 1 with invese gray-scale */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,~i);

/* program output look-up table 1 with inverse gray-scale */
		outportw(base+OUTCSR,olut++);
		o=~i & 0xff;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
	
/* program input look-up table 2 with right-justified 4 MSBs */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i >> 4);

/* program output look-up table 2 to display a right-justified 4-bit image */
		outportw(base+OUTCSR,olut++);
		o=i << 4;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		
/* program input look-up table 3 with left-justified 4 MSBs */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xf0);

/* program output look-up table 3 to display a left-justified 4-bit image */
		outportw(base+OUTCSR,olut++);
		o=i & 0xf0;
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));

/* program input look-up table 4 with left-justified 6 MSBs */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xfc);

/* program output look-up table 4 for addition of two 4-bit images */
		outportw(base+OUTCSR,olut++);
		o1=i >> 4;
		o2=i & 0x0f;
		o=(o1+o2);
		if(o > 0xff) o=0xff;	/* saturation */
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		
/* program input look-up table 5 with left-justified 7 MSBs */
		outportw(base+INCSR1,ilut++);
		outportw(base+INLUT,i & 0xfe);

/* program output look-up table 5 for subtraction of two 4-bit images */
		outportw(base+OUTCSR,olut++);
		o1=i >> 4;
		o2=i & 0x0f;
		o=(o1-o2) & 0xff;
		if(o & 0x10) o=0; /* negative saturation */
		outportw(base+BLUE,o);
		outportw(base+REDGRN,o | (o << 8));
		

/* program input look-up table 6 with bi-level threshold */
		outportw(base+INCSR1,ilut++);
		if(i > 0x80) outportw(base+INLUT,0xff);
		else outportw(base+INLUT,0x00);

/* program output look-up table 6 for display a gray-scale image with color overlay */
		outportw(base+OUTCSR,olut++);
		switch(i & 0x03)
		{
			case 0x00:
			{
				outportw(base+BLUE,i);
				outportw(base+REDGRN,i | (i << 8));
				break;
			}
			case 0x01:
			{	
				outportw(base+BLUE,0);
				outportw(base+REDGRN,REDDATA);
				break;
			}
			case 0x02:
			{
				outportw(base+BLUE,0);
				outportw(base+REDGRN,GREENDATA);
				break;
			}
			default:
			{
				outportw(base+BLUE,0xff);
				outportw(base+REDGRN,0x0000);
			}
		}
		
/* program input look-up table 7 with inverse bi-level threshod */
		outportw(base+INCSR1,ilut++);
		if(i > 0x80) outportw(base+INLUT,0x00);
		else outportw(base+INLUT,0xff);

/* program output look-up table 7 for display a gray-scale image with red overlay */
		outportw(base+OUTCSR,olut++);
		if(i & 1)
		{
			outportw(base+BLUE,0);
			outportw(base+REDGRN,REDDATA);
		}
		else
		{
			outportw(base+BLUE,i);
			outportw(base+REDGRN,i | (i << 8));
		}
	}

	outportw(base+INCSR2,incsr2);
	outportw(base+OUTCSR,outcsr);
	outportw(base+INCSR1,incsr1 & ~(DONEINT | ISEL));

	outportw(base+INCSR2,~(BUFSEL | MODE | WP));
	outportw(base+CURSOR,0x8080);


/* initialize board memory */
	p=(char *)ioremap(basemem[board],2*BUF_SIZE);
	if(p != NULL) memset_io(p,(char)0,2*BUF_SIZE);
	iounmap(p);
}

static loff_t buf_llseek(struct file *file,loff_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= BUF_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static ssize_t buf_read(struct file *file,char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	unsigned char *bufptr;

	if(count < 0) return -EINVAL;

	nbytes=(file->f_pos + count < BUF_SIZE)? count:BUF_SIZE-file->f_pos;

	bufptr=(unsigned char *)file->private_data;

	copy_to_user((char *)buf,bufptr+file->f_pos,nbytes);

	file->f_pos+=nbytes;
	return nbytes;
}

static ssize_t buf_write(struct file *file,const char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	unsigned char *bufptr;

	if(count < 0) return -EINVAL;

	nbytes=(file->f_pos + count < BUF_SIZE)? count:BUF_SIZE-file->f_pos;

	bufptr=(unsigned char *)file->private_data;
	copy_from_user(bufptr+file->f_pos,buf,nbytes);

	file->f_pos+=nbytes;
	return nbytes;
}

static int buf_mmap(struct file *file,struct vm_area_struct *vma)
{
	int board;
	int dev;

	if((vma->vm_end-vma->vm_start+vma->vm_offset) > BUF_SIZE) return -EINVAL;

	board=getboard(file);
	dev=getdev(file);

	vma->vm_offset+=basemem[board]+(dev-BUF0_MINOR)*BUF_SIZE;

	if(vma->vm_offset & ~PAGE_MASK) return -ENXIO;

	if(remap_page_range(vma->vm_start,vma->vm_offset,vma->vm_end-vma->vm_start,vma->vm_page_prot)) return -EAGAIN;

	vma->vm_file=file;
	file->f_dentry->d_inode->i_count++;
	return 0;
}

static int buf_open(struct inode *inode,struct file *file)
{
	int board;
	int dev;
	unsigned char *bufptr;

	MOD_INC_USE_COUNT;

	board=getboard(file);
	dev=getdev(file);

	bufptr=(unsigned char *)ioremap(basemem[board]+(dev-BUF0_MINOR)*BUF_SIZE,BUF_SIZE);
	if(bufptr == NULL)
	{
		MOD_DEC_USE_COUNT;
		return -ENOMEM;
	}
	memset_io(bufptr,(char)0,BUF_SIZE);
	file->private_data=bufptr;
	file->f_pos=0;

	return 0;   
}

static int buf_release(struct inode *inode,struct file *file)
{
	unsigned char *bufptr=(unsigned char *)file->private_data;
	iounmap(bufptr);

	MOD_DEC_USE_COUNT;

	return 0;
}

static struct file_operations buf_fops=
{
	buf_llseek,
	buf_read,	
	buf_write,
	NULL,		/* buf_readdir */
	NULL,		/* buf_pool */
	NULL,		/* buf_ioctl */
	buf_mmap,
	buf_open,
	NULL,		/* buf_flush */	
	buf_release,
	NULL,		/* buf_fsync */
	NULL,		/* buf_fasync */
	NULL,		/* buf_check_media_change */
	NULL,		/* buf_revalidate */
	NULL		/* buf_lock */
};

static loff_t ilut_llseek(struct file *file,loff_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= LUT_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static ssize_t ilut_read(struct file *file,char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	int board;
	int lut;
	int base;
	unsigned short old1;
	unsigned short old2;
	unsigned short lutdata;
	int i;

	if(count < 0) return -EINVAL;

	nbytes=(file->f_pos + count < LUT_SIZE)? count:LUT_SIZE-file->f_pos;

	board=getboard(file);
	lut=getdev(file)-ILUT0_MINOR;
	
	base=baseio[board];

	stop(board);
	old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	old1=inportw(base+INCSR1);
	outportw(base+INCSR1,(old1 & ~ISEL) | (lut & ISEL));
	
	for(i=0;i < nbytes;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		lutdata=inportw(base+INLUT);
		put_user(lutdata,(unsigned char *)buf++);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+INCSR1,old1);	

	return nbytes;
}

static ssize_t ilut_write(struct file *file,const char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	int board;
	int lut;
	int base;
	unsigned short old1;
	unsigned short old2;
	unsigned short lutdata;
	int i;

	if(count < 0) return -EINVAL;

	nbytes=(file->f_pos + count < LUT_SIZE)? count:LUT_SIZE-file->f_pos;

	board=getboard(file);
	lut=getdev(file)-ILUT0_MINOR;
	
	base=baseio[board];

	stop(board);
	old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	old1=inportw(base+INCSR1);
	outportw(base+INCSR1,(old1 & ~ISEL) | (lut & ISEL));
	
	for(i=0;i < nbytes;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		get_user(lutdata,buf++);
		outportw(base+INLUT,lutdata);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+INCSR1,old1);	

	return nbytes;
}

static int ilut_open(struct inode *inode,struct file *file)
{
	int ilut;

	MOD_INC_USE_COUNT;
	ilut=getdev(file)-ILUT0_MINOR;
	clearilut(getboard(file),ilut);

	file->f_pos=0;
	return 0;   
}

static int ilut_release( struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;

	return 0;
}

static struct file_operations ilut_fops=
{

	ilut_llseek,
	ilut_read,	
	ilut_write,
	NULL,		/* ilut_readdir */
	NULL,		/* ilut_pool */
	NULL,		/* ilut_ioctl */
	NULL,		/* ilut_mmap */
	ilut_open,
	NULL,		/* ilut_flush */	
	ilut_release,
	NULL,		/* ilut_fsync */
	NULL,		/* ilut_fasync */
	NULL,		/* ilut_check_media_change */
	NULL,		/* ilut_revalidate */
	NULL		/* ilut_lock */
};

static loff_t olut_llseek(struct file *file,loff_t offset,int orig)
{
	long npos;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= LUT_SIZE)) return -EINVAL;
	file->f_pos=npos;
	return file->f_pos;
}

static ssize_t olut_read(struct file *file,char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	int board;
	int lut;
	int base;
	unsigned short old;
	unsigned short old2;
	unsigned short lutdata;
	int i;

	if((count < 0) || (count % 3)) return -EINVAL;

	nbytes=(file->f_pos + count/3 < LUT_SIZE)? count:3*(LUT_SIZE-file->f_pos);

	board=getboard(file);
	lut=getdev(file)-OLUT0_MINOR;
	
	base=baseio[board];

	stop(board);
	old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,(old & ~OSEL) | (lut & OSEL));
	for(i=0;i < nbytes/3;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		lutdata=inportw(base+REDGRN);
		put_user(lutdata,(unsigned short *) buf);
		buf+=2;
		lutdata=inportw(base+BLUE);
		put_user((unsigned char)lutdata,(unsigned char *)buf++);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+OUTCSR,old);	

	return nbytes;
}

static ssize_t olut_write(struct file *file,const char *buf,size_t count,loff_t *ppos)
{
	int nbytes;
	int board;
	int lut;
	int base;
	unsigned short old;
	unsigned short old2;
	unsigned short lutdata;
	int i;

	if((count < 0) || (count % 3)) return -EINVAL;

	nbytes=(file->f_pos + count/3 < LUT_SIZE)? count:3*(LUT_SIZE-file->f_pos);

	board=getboard(file);
	lut=getdev(file)-OLUT0_MINOR;
	
	 base=baseio[board];

	stop(board);
	old2=inportw(base+INCSR2);
	outportw(base+INCSR2,(old2 & ~MODE) | LOADLUT);
	old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,(old & ~OSEL) | (lut & OSEL));
	
	for(i=0;i < nbytes/3;i++)
	{
		outportw(base+INDEX,file->f_pos++);
		
		get_user(lutdata,(unsigned short *) buf);
		outportw(base+REDGRN,lutdata);
		buf+=2;
		get_user(lutdata,(unsigned short *) buf);
		outportw(base+BLUE,lutdata);
	}
	
	outportw(base+INCSR2,old2);
	outportw(base+OUTCSR,old);	

	return nbytes;
}

static int olut_open(struct inode *inode,struct file *file)
{
	int olut;

	MOD_INC_USE_COUNT;
	
	olut=getdev(file)-OLUT0_MINOR;
	clearolut(getboard(file),olut);
	file->f_pos=0;
	return 0;   
}

static int olut_release( struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;

	return 0;
}

static struct file_operations olut_fops=
{
	olut_llseek,
	olut_read,	
	olut_write,
	NULL,		/* olut_readdir */
	NULL,		/* olut_pool */
	NULL,		/* olut_ioctl */
	NULL,		/* olut_mmap */
	olut_open,
	NULL,		/* olut_flush */	
	olut_release,
	NULL,		/* olut_fsync */
	NULL,		/* olut_fasync */
	NULL,		/* olut_check_media_change */
	NULL,		/* olut_revalidate */
	NULL		/* olut_lock */
};

static loff_t cur_llseek(struct file *file,loff_t offset,int orig)
{
	long npos;
	unsigned short x;
	unsigned short y;
	unsigned short y_2_x_2;
	int board;

	switch(orig)
	{
		case SEEK_SET:
		{
			npos=offset;
			break;
		}
		case SEEK_CUR:
		{
			npos=file->f_pos+offset;
			break;
		}
		case SEEK_END:
		{
			npos=BUF_SIZE-1-offset;
			break;
		}
		default: return -EINVAL;
	}
	if((npos < 0) || (npos >= BUF_SIZE)) return -EINVAL;
	file->f_pos=npos;
	
	x=npos%512;
	y=npos/512;
	
	y_2_x_2=((y/2) << 8) | (x/2);

	board=getboard(file);
	outportw(baseio[board]+CURSOR,y_2_x_2);

	return file->f_pos;
}

static ssize_t cur_read(struct file *file,char *buf,size_t count,loff_t *ppos)
{
	int board;
	unsigned short cursor;
	int x;
	int y;

	if(count != 2*sizeof(int)) return -EINVAL;

	board=getboard(file);
	cursor=inportw(baseio[board]+CURSOR);
	x=(cursor & 0xff)*2;
	put_user(x,(int *)buf);
	y=(cursor >> 8)*2;
	put_user(y,(int *)(buf+sizeof(int)));
	return 2*sizeof(int);
}

static ssize_t cur_write(struct file *file,const char *buf,size_t count,loff_t *ppos)
{
	unsigned short x;
	unsigned short y;
	unsigned short y_2_x_2;
	int board;

	if(count != 2*sizeof(int)) return -EINVAL;

	get_user(x,(int *)buf);
	get_user(y,(int *)(buf+sizeof(int)));
	
	y_2_x_2=((y/2) << 8) | (x/2);

	board=getboard(file);
	outportw(baseio[board]+CURSOR,y_2_x_2);

	return 2*sizeof(int);
}

static int cur_ioctl(struct inode *inode,struct file *file,unsigned int cmd,unsigned long arg)
{
	int board;
	unsigned int curport;
	unsigned short old;

	board=getboard(file);
	curport=baseio[board]+OUTCSR;
	old=inportw(curport);
	switch(cmd)
	{
		case CUR_ON:
		{
			outportw(curport,old | CURS);
			break;
		}
		case CUR_OFF:
		{
			outportw(curport,old & ~CURS);
			break;
		}
		case CUR_QUERY:	
		{
			int status=(inportw(curport) & CURS) >> 6;
			put_user(status,(int *)arg);
			break;
		}
		default: return -EINVAL;
	}
	return 0;
}

static int cur_open(struct inode *inode,struct file *file)
{
	int board;
	unsigned int base;
	unsigned short old;

	MOD_INC_USE_COUNT;
	
	file->f_pos=BUF_SIZE/2;
	board=getboard(file);
	base=baseio[board];
	old=inportw(base+OUTCSR);
	outportw(base+OUTCSR,old & ~CURS);
	outportw(base+CURSOR,0x8080);

	return 0;   
}

static int cur_release(struct inode *inode,struct file *file)
{
	int board;
	unsigned int curport;
	unsigned short old;

	board=getboard(file);
	curport=baseio[board]+OUTCSR;
	old=inportw(curport);
	outportw(curport,old & ~CURS);

	MOD_DEC_USE_COUNT;

	return 0;
}

static struct file_operations cur_fops=
{
	cur_llseek,
	cur_read,	
	cur_write,
	NULL,		/* cur_readdir */
	NULL,		/* cur_pool */
	cur_ioctl,
	NULL,		/* cur_mmap */
	cur_open,
	NULL,		/* cur_flush */	
	cur_release,
	NULL,		/* cur_fsync */
	NULL,		/* cur_fasync */
	NULL,		/* cur_check_media_change */
	NULL,		/* cur_revalidate */
	NULL		/* cur_lock */
};

static loff_t cs_llseek(struct file *file,loff_t offset,int orig)
{
	return -EINVAL;
}

static int cs_ioctl(struct inode *inode,struct file *file,unsigned int cmd,unsigned long arg)
{
	int board;
	unsigned int base;
	int mode;
	unsigned short incsr1;
	unsigned short incsr2;
	unsigned short outcsr=0;
	int doneint;
	int syncint;
	int lut;
	int buf;
	int prot;
	int busy;
	int opmode; 
	int timing;
	int display;
	int field;
	int vsync;
	int frproc;
	int exttrg;

	board=getboard(file);
	base=baseio[board];
	switch(cmd)
	{
		case CS_CLEAR:
		{
			clear(board);
			break;
		}
		case CS_SETMODE:
		{
			stop(board);
			mode=(arg << 4) & MODE;
			incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~MODE) | mode);
			break;
		}
		case CS_MODE:
		{
			mode=(inportw(base+INCSR2) & MODE) >> 4;
			put_user(mode,(int *)arg);
			break;
		}
		case CS_ENABLEINT:
		{
			if(!request_irq(irq_data[board].irq,irq_handler,0,"DT2851",NULL))
			{
				irq_data[board].p=current;
				enable_irq(irq_data[board].irq);

				incsr1=inportw(base+INCSR1);
				doneint=(arg << 6) & DONEINT;
				outportw(base+INCSR1,(incsr1 & ~DONEINT) | doneint);

				outcsr=inportw(base+OUTCSR);
				syncint=(arg << (14-1)) & SYNCINT;
				outportw(base+OUTCSR,(outcsr & ~SYNCINT) | syncint);
			}
			else return -EACCES;
			break;
		}
		case CS_DISABLEINT:
		{
			incsr1=inportw(base+INCSR1);
			doneint=(arg << 6) & DONEINT;
			outportw(base+INCSR1,(incsr1 & ~DONEINT) & ~doneint);

			outcsr=inportw(base+OUTCSR);
			syncint=(arg << (14-1)) & SYNCINT;
			outportw(base+OUTCSR,(outcsr & ~SYNCINT) & ~syncint);

			disable_irq(irq_data[board].irq);
			irq_data[board].p=NULL;
			free_irq(irq_data[board].irq,NULL);
			break;
		}
		case CS_INTACK:
		{
			incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~DONEINT));
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~SYNCINT));

			outportw(base+INCSR1,incsr1);	
			outportw(base+OUTCSR,outcsr);
			break;
		}
		case CS_SELILUT:
		{
			lut=arg & ISEL;
			incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~ISEL) | lut);
			break;
		}
		case CS_ILUT:
		{
			lut=inportw(base+INCSR1) & ISEL;
			put_user(lut,(int *)arg);
			break;
		}
		case CS_SELOLUT:
		{
			lut=arg & OSEL;
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~OSEL) | lut);
			break;
		}
		case CS_OLUT:
		{
			lut=inportw(base+OUTCSR) & OSEL;
			put_user(lut,(int *)arg);
			break;
		}
		case CS_SELIBUF:
		{
			stop(board);
			buf=(arg << 7) & BUFSEL;
			incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~BUFSEL) | buf);
			break;
		}
		case CS_IBUF:
		{
			buf=(inportw(base+INCSR2) & BUFSEL) >> 7;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_SELOBUF:
		{
			buf=(arg << 4) & DISBUF;
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~DISBUF) | buf);
			break;
		}
		case CS_OBUF:
		{
			buf=(inportw(base+OUTCSR) & DISBUF) >> 4;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_SETPROT:
		{
			stop(board);
			prot=arg & WP;
			incsr2=inportw(base+INCSR2);
			outportw(base+INCSR2,(incsr2 & ~WP) | prot);
			break;
		}
		case CS_PROT:
		{
			buf=inportw(base+INCSR2) & WP;
			put_user(buf,(int *)arg);
			break;
		}
		case CS_START:
		{
			stop(board);
			opmode=((~arg) << 3) & ENSTOP;
			incsr1=inportw(base+INCSR1);
			outportw(base+INCSR1,(incsr1 & ~ENSTOP) | opmode);
			outportw(base+INCSR1,(incsr1 & ~ENSTOP) | opmode | BUSY);
			break;
		}
		case CS_STOP:
		{
			stop(board);
			break;
		}
		case CS_BUSY:
		{
			busy=(inportw(base+INCSR1) & BUSY) >> 7;
			put_user(busy,(int *)arg);
			break;
		}
		case CS_DISPLAYON:
		{
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,outcsr | DISP);
			break;
		}
		case CS_DISPLAYOFF:
		{
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,outcsr & ~DISP);
			break;
		}
		case CS_SETTIMING:
		{
			timing=(arg << 5) & EXTTIM;
			outcsr=inportw(base+OUTCSR);
			outportw(base+OUTCSR,(outcsr & ~EXTTIM) | timing);
			break;
		}
		case CS_TIMING:
		{
			timing=(inportw(base+OUTCSR) & EXTTIM) >> 5;
			put_user(timing,(int *)arg);
			break;
		}
		case CS_DISPLAY:
		{
			display=(inportw(base+OUTCSR) & DISP) >> 7;
			put_user(display,(int *)arg);
			break;
		}
		case CS_FIELD:
		{
			field=(inportw(base+OUTCSR) & FIELD) >> 13;
			put_user(field,(int *)arg);
			break;
		}
		case CS_VSYNC:
		{
			vsync=(inportw(base+OUTCSR) & VSYNC) >> 15;
			put_user(vsync,(int *)arg);
			break;
		}
		case CS_FRPROC:
		{
			frproc=(~inportw(base+INCSR1) & FRPROC) >> 4;
			put_user(frproc,(int *)arg);
			break;
		}
		case CS_EXTTRG:
		{
			exttrg=(inportw(base+INCSR1) & EXTTRG) >> 5;
			put_user(exttrg,(int *)arg);
			break;
		}

		default: return -EINVAL;
	}
	return 0;
}

static int cs_open(struct inode *inode,struct file *file)
{
	int board;
	unsigned int base;
	unsigned short outcsr;
	unsigned short incsr1;

	MOD_INC_USE_COUNT;

	board=getboard(file);
	base=baseio[board];

	outcsr=inportw(base+OUTCSR);
	outportw(base+OUTCSR,outcsr & ~(SYNCINT | DISP | CURS | EXTTIM | DISBUF | OSEL));

	incsr1=inportw(base+INCSR1);
	incsr1|=ENSTOP;
	outportw(base+INCSR1,incsr1);
	outportw(base+INCSR1,incsr1 & ~BUSY);

	outportw(base+INCSR2,~(BUFSEL | MODE | WP));

	return 0;   

}

static int cs_release(struct inode *inode,struct file *file)
{
	MOD_DEC_USE_COUNT;

	return 0;
}

static struct file_operations cs_fops=
{
	cs_llseek,
	NULL,		/* cs_read */
	NULL,		/* cs_write */
	NULL,		/* cs_readdir */
	NULL,		/* cs_pool */
	cs_ioctl,
	NULL,		/* cs_mmap */
	cs_open,
	NULL,		/* cs_flush */	
	cs_release,
	NULL,		/* cs_fsync */
	NULL,		/* cs_fasync */
	NULL,		/* cs_check_media_change */
	NULL,		/* cs_revalidate */
	NULL		/* cs_lock */
};


static int dt_open(struct inode *inode,struct file *file)
{
	int minor;
	int i;

	minor=MINOR(file->f_dentry->d_inode->i_rdev);
	for(i=1;i < 4;i++)
	{
		if(minor > CS_MINOR) minor-=CS_MINOR+1;
	}

	if((minor >= BUF0_MINOR) && (minor <= BUF1_MINOR)) file->f_op= &buf_fops;
	else if((minor >= ILUT0_MINOR) && (minor <= ILUT7_MINOR)) file->f_op = &ilut_fops;
	else if((minor >= OLUT0_MINOR) && (minor <= OLUT7_MINOR)) file->f_op = &olut_fops;
	else if(minor == CUR_MINOR) file->f_op=&cur_fops;
	else if(minor == CS_MINOR) file->f_op=&cs_fops;
	else return -ENXIO;

	if(file->f_op && file->f_op->open) return file->f_op->open(inode,file);
	return 0;   
}


static struct file_operations dt_fops=
{
	NULL,		/* dt_llseek */
	NULL,		/* dt_read */
	NULL,		/* dt_write */
	NULL,		/* dt_readdir */
	NULL,		/* dt_pool */
	NULL,		/* dt_ioctl */
	NULL,		/* dt_mmap */
	dt_open,
	NULL,		/* dt_flush */	
	NULL,		/* dt_release */
	NULL,		/* dt_fsync */
	NULL,		/* dt_fasync */
	NULL,		/* dt_check_media_change */
	NULL,		/* dt_revalidate */
	NULL		/* dt_lock */
};

int init_module(void)
{
	int err;
	int i;

	printk("DT2851: ");
	printk("Copyright (C) 1997...1999 Walter Fetter Lages <w.fetter@ieee.org>.\n");
	
	for(i=0;i < 4;i++)
	{
		irq_data[i].irq=irq[i];
		irq_data[i].p=NULL;
	}

	if(register_chrdev(major,"DT2851",&dt_fops))
	{
		printk("major %d can't be registered.\n",major);
		return -EIO;
	}

	err=0;

	if(check_region(baseio[0],DT_IO_EXTENT))
	{
		printk("I/O ports are already allocated to another device.\n");
		err=-EIO;
	}
	request_region(baseio[0],DT_IO_EXTENT,"DT2851");

	for(i=1;i < 4;i++)
	{	
		if(baseio[i]!=baseio[i-1])
		{
			if(check_region(baseio[i],DT_IO_EXTENT))
			{
				printk("I/O ports are already allocated to another device.\n");
				err=-EIO;
			}
			request_region(baseio[i],DT_IO_EXTENT,"DT2851");
		}
	}
	if(err) for(i=0;i < 4;i++) release_region(baseio[i],DT_IO_EXTENT);

	return err;
}

void cleanup_module(void)
{
	int i;

	for(i=0;i < 4;i++) release_region(baseio[i],DT_IO_EXTENT);
	if(unregister_chrdev(major,"DT2851") != 0) printk("major %d can't be unregistered\n",major);
}

